@@ -10,16 +10,30 @@
 #include <config.h>
 #endif
 
+#include <glib/gerror.h>
+#include <gtk/gtkmain.h>
+#include <gtk/gtkrange.h>
 #include <gtk/gtkbutton.h>
 #include <gtk/gtkoptionmenu.h>
-#include <glib/gerror.h>
 
 #include <libgnome/gnome-i18n.h>
+#include <libgnomeui/gnome-about.h>
 
+#include <bonobo/bonobo-exception.h>
+#include <bonobo/bonobo-control.h>
+#include <bonobo/bonobo-widget.h>
+
+#include <bonobo-activation/bonobo-activation.h>
+
+#include "cddb.h"
 #include "cdrom.h"
 #include "display.h"
 #include "gnome-cd.h"
 
+#include "GNOME_Media_CDDBSlave2.h"
+
+static GNOME_Media_CDDBTrackEditor track_editor = CORBA_OBJECT_NIL;
+
 static void
 maybe_close_tray (GnomeCD *gcd)
 {
@@ -27,7 +41,7 @@ maybe_close_tray (GnomeCD *gcd)
 	GError *error;
 
 	if (gnome_cdrom_get_status (gcd->cdrom, &status, &error) == FALSE) {
-		g_warning ("%s: %s", __FUNCTION__, error->message);
+		gcd_warning ("%s", error);
 		g_error_free (error);
 		
 		return;
@@ -35,7 +49,7 @@ maybe_close_tray (GnomeCD *gcd)
 
 	if (status->cd == GNOME_CDROM_STATUS_TRAY_OPEN) {
 		if (gnome_cdrom_close_tray (gcd->cdrom, &error) == FALSE) {
-			g_warning ("%s: %s", __FUNCTION__, error->message);
+			gcd_warning ("%s", error);
 			g_error_free (error);
 		}
 	}
@@ -51,7 +65,7 @@ eject_cb (GtkButton *button,
 	GError *error;
 
 	if (gnome_cdrom_eject (gcd->cdrom, &error) == FALSE) {
-		g_warning ("%s: %s", __FUNCTION__, error->message);
+		gcd_warning ("%s", error);
 		g_error_free (error);
 	}
 
@@ -79,7 +93,7 @@ play_cb (GtkButton *button,
 	GnomeCDRomMSF *endmsf;
 
 	if (gnome_cdrom_get_status (gcd->cdrom, &status, &error) == FALSE) {
-		g_warning ("%s: %s", __FUNCTION__, error->message);
+		gcd_warning ("%s", error);
 		g_error_free (error);
 		return;
 
@@ -88,7 +102,7 @@ play_cb (GtkButton *button,
 	switch (status->cd) {
 	case GNOME_CDROM_STATUS_TRAY_OPEN:
 		if (gnome_cdrom_close_tray (gcd->cdrom, &error) == FALSE) {
-			g_warning ("Cannot close tray: %s", error->message);
+			gcd_warning ("Cannot close tray: %s", error);
 			g_error_free (error);
 
 			g_free (status);
@@ -103,12 +117,12 @@ play_cb (GtkButton *button,
 			end_track = -1;
 			endmsf = NULL;
 		} else {
-			end_track == 2;
+			end_track = 2;
 			endmsf = &msf;
 		}
 		
 		if (gnome_cdrom_play (gcd->cdrom, 1, &msf, end_track, endmsf, &error) == FALSE) {
-			g_warning ("%s: %s", __FUNCTION__, error->message);
+			g_warning ("%s: %s", G_GNUC_FUNCTION, error->message);
 			g_error_free (error);
 
 			g_free (status);
@@ -117,7 +131,7 @@ play_cb (GtkButton *button,
 		break;
 
 	case GNOME_CDROM_STATUS_DRIVE_NOT_READY:
-		g_warning ("Drive not ready");
+		gcd_warning ("Drive not ready: %s", NULL);
 
 		g_free (status);
 		return;
@@ -129,20 +143,29 @@ play_cb (GtkButton *button,
 
 	switch (status->audio) {
 	case GNOME_CDROM_AUDIO_PLAY:
-		if (gnome_cdrom_pause (gcd->cdrom, &error) == FALSE) {
-			g_warning ("%s: %s", __FUNCTION__, error->message);
-			g_error_free (error);
-
-			g_free (status);
-			return;
-		}
-
-		if (gcd->current_image == gcd->pause_image) {
+		if (gcd->current_image == gcd->play_image) {
 			aob = gtk_widget_get_accessible (GTK_WIDGET (gcd->play_b));
-			atk_object_set_name (aob, _("Play"));
+			atk_object_set_name (aob, _("Pause"));
 			gtk_container_remove (GTK_CONTAINER (gcd->play_b), gcd->current_image);
-			gtk_container_add (GTK_CONTAINER (gcd->play_b), gcd->play_image);
-			gcd->current_image = gcd->play_image;
+			gtk_container_add (GTK_CONTAINER (gcd->play_b), gcd->pause_image);
+			gcd->current_image = gcd->pause_image;
+		} else { 
+			if (gnome_cdrom_pause (gcd->cdrom, &error) == FALSE) {
+				gcd_warning ("%s", error);
+				g_error_free (error);
+
+				g_free (status);
+				return;
+			}
+
+			if (gcd->current_image == gcd->pause_image) {
+				aob = gtk_widget_get_accessible (GTK_WIDGET (gcd->play_b));
+				atk_object_set_name (aob, _("Play"));
+				gtk_container_remove (GTK_CONTAINER (gcd->play_b), gcd->current_image);
+				gtk_container_add (GTK_CONTAINER (gcd->play_b),
+						   gcd->play_image);
+				gcd->current_image = gcd->play_image;
+			}
 		}
 		break;
 
@@ -159,7 +182,7 @@ play_cb (GtkButton *button,
 		}
 		if (gnome_cdrom_play (gcd->cdrom, status->track,
 				      &status->relative, end_track, endmsf, &error) == FALSE) {
-			g_warning ("%s: %s", __FUNCTION__, error->message);
+			gcd_warning ("%s", error);
 			g_error_free (error);
 			
 			g_free (status);
@@ -189,7 +212,7 @@ play_cb (GtkButton *button,
 			endmsf = &msf;
 		}
 		if (gnome_cdrom_play (gcd->cdrom, 1, &msf, end_track, endmsf, &error) == FALSE) {
-			g_warning ("%s: %s", __FUNCTION__, error->message);
+			gcd_warning ("%s", error);
 			g_error_free (error);
 
 			g_free (status);
@@ -206,7 +229,7 @@ play_cb (GtkButton *button,
 		break;
 
 	case GNOME_CDROM_AUDIO_ERROR:
-		g_warning ("Error playing CD");
+		gcd_warning ("Error playing CD: %s", NULL);
 		g_free (status);
 		return;
 
@@ -228,7 +251,7 @@ stop_cb (GtkButton *button,
 	maybe_close_tray (gcd);
 
 	if (gnome_cdrom_stop (gcd->cdrom, &error) == FALSE) {
-		g_warning ("%s: %s", __FUNCTION__, error->message);
+		gcd_warning ("%s", error);
 		g_error_free (error);
 	}
 
@@ -251,7 +274,7 @@ ffwd_timeout_cb (gpointer data)
 	GnomeCD *gcd = data;
 
 	if (gnome_cdrom_fast_forward (gcd->cdrom, &error) == FALSE) {
-		g_warning ("%s: %s", __FUNCTION__, error->message);
+		gcd_warning ("%s", error);
 		g_error_free (error);
 		return FALSE;
 	}
@@ -264,8 +287,14 @@ ffwd_press_cb (GtkButton *button,
 	       GdkEvent *ev,
 	       GnomeCD *gcd)
 {
+	if (gcd->timeout > 0) {
+		return FALSE;
+	}
+	
 	maybe_close_tray (gcd);
 
+	ffwd_timeout_cb (gcd);
+	
 	gcd->timeout = gtk_timeout_add (140, ffwd_timeout_cb, gcd);
 
 	if (gcd->current_image == gcd->play_image) {
@@ -289,6 +318,7 @@ ffwd_release_cb (GtkButton *button,
 {
 	if (gcd->timeout > 0) {
 		gtk_timeout_remove (gcd->timeout);
+		gcd->timeout = 0;
 	}
 
 	return FALSE;
@@ -303,7 +333,7 @@ next_cb (GtkButton *button,
 	maybe_close_tray (gcd);
 
 	if (gnome_cdrom_next (gcd->cdrom, &error) == FALSE) {
-		g_warning ("%s: %s", __FUNCTION__, error->message);
+		gcd_warning ("%s", error);
 		g_error_free (error);
 	}
 
@@ -327,7 +357,7 @@ back_cb (GtkButton *button,
 	maybe_close_tray (gcd);
 
 	if (gnome_cdrom_back (gcd->cdrom, &error) == FALSE) {
-		g_warning ("%s: %s", __FUNCTION__, error->message);
+		gcd_warning ("%s", error);
 		g_error_free (error);
 	}
 
@@ -349,7 +379,7 @@ rewind_timeout_cb (gpointer data)
 	GnomeCD *gcd = data;
 	
 	if (gnome_cdrom_rewind (gcd->cdrom, &error) == FALSE) {
-		g_warning ("%s: %s", __FUNCTION__, error->message);
+		gcd_warning ("%s", error);
 		g_error_free (error);
 		return FALSE;
 	}
@@ -362,8 +392,14 @@ rewind_press_cb (GtkButton *button,
 		 GdkEvent *ev,
 		 GnomeCD *gcd)
 {
+	if (gcd->timeout > 0) {
+		return FALSE;
+	}
+	
 	maybe_close_tray (gcd);
 
+	/* Call it so a click will activate it */
+	rewind_timeout_cb (gcd);
 	gcd->timeout = gtk_timeout_add (140, rewind_timeout_cb, gcd);
 
 	if (gcd->current_image == gcd->play_image) {
@@ -386,32 +422,233 @@ rewind_release_cb (GtkButton *button,
 {
 	if (gcd->timeout > 0) {
 		gtk_timeout_remove (gcd->timeout);
+		gcd->timeout = 0;
 	}
 	
 	return FALSE;
 }
-	
-void
-mixer_cb (GtkButton *button,
-	  GnomeCD *gcd)
+
+static void
+set_track_option_menu (GtkOptionMenu *menu,
+		       int track)
 {
-	char *mixer_path;
-	char *argv[2] = {NULL, NULL};
-	GError *error = NULL;
-	gboolean ret;
+	g_signal_handlers_block_matched (G_OBJECT (menu),
+					 G_SIGNAL_MATCH_FUNC,
+					 0, 0, NULL,
+					 G_CALLBACK (skip_to_track), NULL);
+	gtk_option_menu_set_history (menu, track - 1);
+	g_signal_handlers_unblock_matched (G_OBJECT (menu),
+					   G_SIGNAL_MATCH_FUNC,
+					   0, 0, NULL,
+					   G_CALLBACK (skip_to_track), NULL);
+}
 
-	/* Open the mixer */
-	mixer_path = g_find_program_in_path ("gmix");
+static void
+set_window_track_title (GnomeCD *gcd,
+			GnomeCDRomStatus *status)
+{
+	int idx = status->track - 1;
+	const char *artist = "";
+	const char *track_name = "";
 
-	argv[0] = mixer_path;
-	ret = g_spawn_async (NULL, argv, NULL, 0, NULL, NULL, NULL, &error);
-	if (ret == FALSE) {
-		g_warning ("There was an error starting %s: %s", mixer_path,
-			   error->message);
-		g_error_free (error);
+	if (gcd->disc_info) {
+		if (idx >= 0 && idx < gcd->disc_info->ntracks)
+			track_name = gcd->disc_info->track_info [idx]->name;
+
+		artist = gcd->disc_info->artist ? gcd->disc_info->artist : "";
 	}
 
-	g_free (mixer_path);
+	gnome_cd_set_window_title (gcd, artist, track_name);
+}
+
+/* Do all the stuff for when the status is ok */
+static void
+status_ok (GnomeCD *gcd,
+	   GnomeCDRomStatus *status)
+{
+	AtkObject *aob;
+	int track;
+	char *text;
+	
+	/* Allow the track editor to work */
+	gtk_widget_set_sensitive (gcd->trackeditor_b, TRUE);
+
+	/* Set the track menu on. */
+	gtk_widget_set_sensitive (gcd->tracks, TRUE);
+	
+	/* All buttons can be used when the tray is open,
+	   they should shut the tray. */
+	gtk_widget_set_sensitive (gcd->back_b, TRUE);
+	gtk_widget_set_sensitive (gcd->rewind_b, TRUE);
+	gtk_widget_set_sensitive (gcd->play_b, TRUE);
+	gtk_widget_set_sensitive (gcd->stop_b, TRUE);
+	gtk_widget_set_sensitive (gcd->ffwd_b, TRUE);
+	gtk_widget_set_sensitive (gcd->next_b, TRUE);
+	gtk_widget_set_sensitive (gcd->eject_b, TRUE);
+
+	switch (status->audio) {
+	case GNOME_CDROM_AUDIO_NOTHING:
+		break;
+
+	case GNOME_CDROM_AUDIO_PLAY:
+		/* Change the play button to pause */
+		aob = gtk_widget_get_accessible (GTK_WIDGET (gcd->play_b));
+		atk_object_set_name (aob, _("Pause"));
+
+		if (gcd->current_image != gcd->pause_image) {
+			gtk_container_remove (GTK_CONTAINER (gcd->play_b),
+					      gcd->play_image);
+			gtk_container_add (GTK_CONTAINER (gcd->play_b),
+					   gcd->pause_image);
+			gcd->current_image = gcd->pause_image;
+		}
+		/* Find out if the track has changed */
+		track = gtk_option_menu_get_history (GTK_OPTION_MENU (gcd->tracks));
+		if (track + 1 != status->track) {
+			set_track_option_menu (GTK_OPTION_MENU (gcd->tracks),
+					       status->track);
+		}
+		
+		
+/*  		cd_display_clear (CD_DISPLAY (gcd->display)); */
+		if (status->relative.second >= 10) {
+			text = g_strdup_printf ("%d:%d", status->relative.minute, status->relative.second);
+		} else {
+			text = g_strdup_printf ("%d:0%d", status->relative.minute, status->relative.second);
+		}
+		
+		cd_display_set_line (CD_DISPLAY (gcd->display),
+				     CD_DISPLAY_LINE_TIME, text);
+		g_free (text);
+		set_window_track_title (gcd, status);
+			
+		break;
+
+	case GNOME_CDROM_AUDIO_PAUSE:
+		/* Change the play button to pause */
+		aob = gtk_widget_get_accessible (GTK_WIDGET (gcd->play_b));
+		atk_object_set_name (aob, _("Play"));
+		if (gcd->current_image != gcd->play_image) {
+			gtk_container_remove (GTK_CONTAINER (gcd->play_b),
+					      gcd->pause_image);
+			gtk_container_add (GTK_CONTAINER (gcd->play_b),
+					   gcd->play_image);
+			gcd->current_image = gcd->play_image;
+		}
+
+		set_window_track_title (gcd, status);
+		break;
+		
+	case GNOME_CDROM_AUDIO_COMPLETE:
+		if (gcd->cdrom->loopmode == GNOME_CDROM_LOOP) {
+			if (gcd->cdrom->playmode == GNOME_CDROM_WHOLE_CD) {
+				/* Around we go */
+				play_cb (NULL, gcd);
+			} else {
+				GnomeCDRomMSF msf;
+				int start_track, end_track;
+				GError *error;
+				
+				/* CD has gone to the start of the next track.
+				   Track we want to loop is track - 1 */
+				start_track = status->track - 1;
+				end_track = status->track;
+				
+				msf.minute = 0;
+				msf.second = 0;
+				msf.frame = 0;
+				
+				if (gnome_cdrom_play (gcd->cdrom, start_track, &msf,
+						      end_track, &msf, &error) == FALSE) {
+					gcd_warning ("%s", error);
+					g_error_free (error);
+					
+					g_free (status);
+					return;
+				}
+			}
+		} else {
+			/* We've stopped playing anything
+			   reset play button, and track option menu */
+			if (gcd->current_image != gcd->play_image) {
+				gtk_container_remove (GTK_CONTAINER (gcd->play_b),
+						      gcd->pause_image);
+				gtk_container_add (GTK_CONTAINER (gcd->play_b),
+						   gcd->play_image);
+				gcd->current_image = gcd->play_image;
+			}
+
+			set_track_option_menu (GTK_OPTION_MENU (gcd->tracks), 1);
+		}		
+		break;
+		
+	case GNOME_CDROM_AUDIO_STOP:
+		cd_display_set_line (CD_DISPLAY (gcd->display),
+				     CD_DISPLAY_LINE_TIME, "");
+		if (gcd->disc_info != NULL) {
+			gnome_cd_set_window_title (gcd,
+						   gcd->disc_info->artist ?
+						   gcd->disc_info->artist : "",
+						   gcd->disc_info->title ?
+						   gcd->disc_info->title : "");
+		} else {
+			gnome_cd_set_window_title (gcd, NULL, NULL);
+		}
+		break;
+		
+	case GNOME_CDROM_AUDIO_ERROR:
+		cd_display_clear (CD_DISPLAY (gcd->display));
+		cd_display_set_line (CD_DISPLAY (gcd->display),
+				     CD_DISPLAY_LINE_TIME, _("Disc error"));
+		if (gcd->disc_info != NULL) {
+			gnome_cd_set_window_title (gcd, gcd->disc_info->artist ?
+						   gcd->disc_info->artist : "",
+						   gcd->disc_info->title ?
+						   gcd->disc_info->title : "");
+		} else {
+			gnome_cd_set_window_title (gcd, NULL, NULL);
+		}
+		break;
+		
+	default:
+		if (gcd->disc_info != NULL) {
+			gnome_cd_set_window_title (gcd,
+						   gcd->disc_info->artist ?
+						   gcd->disc_info->artist : "",
+						   gcd->disc_info->title ?
+						   gcd->disc_info->title : "");
+		} else {
+			gnome_cd_set_window_title (gcd, NULL, NULL);
+		}
+		
+		break;
+	}
+	
+	if (gcd->last_status == NULL ||
+	    gcd->last_status->cd != GNOME_CDROM_STATUS_OK) {
+		cddb_get_query (gcd);
+	}
+		
+	if (gcd->disc_info == NULL) {
+		cd_display_set_line (CD_DISPLAY (gcd->display),
+				     CD_DISPLAY_LINE_ARTIST,
+				     _("Unknown Artist"));
+		cd_display_set_line (CD_DISPLAY (gcd->display),
+				     CD_DISPLAY_LINE_ALBUM,
+				     _("Unknown Album"));
+		gnome_cd_set_window_title (gcd, NULL, NULL);
+	} else {
+		GnomeCDDiscInfo *info = gcd->disc_info;
+		
+		cd_display_set_line (CD_DISPLAY (gcd->display),
+				     CD_DISPLAY_LINE_ARTIST,
+				     info->artist ? info->artist :
+				     _("Unknown Artist"));
+		cd_display_set_line (CD_DISPLAY (gcd->display),
+				     CD_DISPLAY_LINE_ALBUM,
+				     info->title ? info->title :
+				     _("Unknown Album"));
+	}
 }
 
 void
@@ -419,118 +656,79 @@ cd_status_changed_cb (GnomeCDRom *cdrom,
 		      GnomeCDRomStatus *status,
 		      GnomeCD *gcd)
 {
-	char *text;
-	int track;
+	if (gcd->not_ready == TRUE) {
+		return;
+	}
 	
 	switch (status->cd) {
 	case GNOME_CDROM_STATUS_OK:
-		if (status->audio == GNOME_CDROM_AUDIO_COMPLETE) {
-			if (cdrom->loopmode == GNOME_CDROM_LOOP) {
-				if (cdrom->playmode == GNOME_CDROM_WHOLE_CD) {
-					/* Around we go */
-					play_cb (NULL, gcd);
-				} else {
-					GnomeCDRomMSF msf;
-					int start_track, end_track;
-					GError *error;
-
-					/* CD has gone to the start of the next track.
-					   Track we want to loop is track - 1 */
-					start_track = status->track - 1;
-					end_track = status->track;
-
-					msf.minute = 0;
-					msf.second = 0;
-					msf.frame = 0;
-					
-					if (gnome_cdrom_play (gcd->cdrom, start_track, &msf,
-							      end_track, &msf, &error) == FALSE) {
-						g_warning ("%s: %s", __FUNCTION__, error->message);
-						g_error_free (error);
-						
-						g_free (status);
-						return;
-					}
-				}
-			}
-			break;
-		}
-		
-		track = gtk_option_menu_get_history (GTK_OPTION_MENU (gcd->tracks));
-		if (track + 1 != status->track) {
-			g_signal_handlers_block_matched (G_OBJECT (gcd->tracks), G_SIGNAL_MATCH_FUNC,
-							 0, 0, NULL, G_CALLBACK (skip_to_track), gcd);
-			gtk_option_menu_set_history (GTK_OPTION_MENU (gcd->tracks), status->track - 1);
-			g_signal_handlers_unblock_matched (G_OBJECT (gcd->tracks), G_SIGNAL_MATCH_FUNC,
-							   0, 0, NULL, G_CALLBACK (skip_to_track), gcd);
-		}
-		
-		if (gcd->last_status == NULL || gcd->last_status->cd != GNOME_CDROM_STATUS_OK) {
-			cddb_get_query (gcd);
-		}
-
-		cd_display_clear (CD_DISPLAY (gcd->display));
-		if (status->relative.second >= 10) {
-			text = g_strdup_printf ("%d:%d", status->relative.minute, status->relative.second);
-		} else {
-			text = g_strdup_printf ("%d:0%d", status->relative.minute, status->relative.second);
-		}
-
-		cd_display_set_line (CD_DISPLAY (gcd->display), CD_DISPLAY_LINE_TIME, text);
-		g_free (text);
-
-		if (gcd->disc_info == NULL) {
-			cd_display_set_line (CD_DISPLAY (gcd->display), CD_DISPLAY_LINE_ARTIST, "Unknown Artist");
-			cd_display_set_line (CD_DISPLAY (gcd->display), CD_DISPLAY_LINE_ALBUM, "Unknown Album");
-			gnome_cd_set_window_title (gcd, NULL, NULL);
-		} else {
-			GnomeCDDiscInfo *info = gcd->disc_info;
-
-			if (info->artist != NULL) {
-				cd_display_set_line (CD_DISPLAY (gcd->display), CD_DISPLAY_LINE_ARTIST, info->artist);
-			}
-			if (info->title != NULL) {
-				cd_display_set_line (CD_DISPLAY (gcd->display), CD_DISPLAY_LINE_ALBUM, info->title);
-			}
-
-			if (status->audio == GNOME_CDROM_AUDIO_PLAY ||
-			    status->audio == GNOME_CDROM_AUDIO_PAUSE) {
-				gnome_cd_set_window_title (gcd, info->artist, info->tracknames[status->track - 1]);
-			} else {
-				gnome_cd_set_window_title (gcd, info->artist, info->title);
-			}
-		}
+		status_ok (gcd, status);
 		break;
-
 		
 	case GNOME_CDROM_STATUS_NO_DISC:
+		if (gcd->disc_info != NULL) {
+			cddb_free_disc_info (gcd->disc_info);
+			gcd->disc_info = NULL;
+		}
+		
+		gtk_widget_set_sensitive (gcd->trackeditor_b, FALSE);
+
+		/* Set the other widgets off */
+		gtk_widget_set_sensitive (gcd->tracks, FALSE);
+		gtk_widget_set_sensitive (gcd->back_b, FALSE);
+		gtk_widget_set_sensitive (gcd->rewind_b, FALSE);
+		gtk_widget_set_sensitive (gcd->play_b, FALSE);
+		gtk_widget_set_sensitive (gcd->stop_b, FALSE);
+		gtk_widget_set_sensitive (gcd->ffwd_b, FALSE);
+		gtk_widget_set_sensitive (gcd->next_b, FALSE);
+
+		/* We can still use the eject button when there's no disc. */
+		gtk_widget_set_sensitive (gcd->eject_b, TRUE);
+
 		cd_display_clear (CD_DISPLAY (gcd->display));
-		cd_display_set_line (CD_DISPLAY (gcd->display), CD_DISPLAY_LINE_TIME, "No disc");
+		cd_display_set_line (CD_DISPLAY (gcd->display), CD_DISPLAY_LINE_TIME, _("No disc"));
 		gnome_cd_set_window_title (gcd, NULL, NULL);
 		break;
 
 	case GNOME_CDROM_STATUS_TRAY_OPEN:
+		if (gcd->disc_info != NULL) {
+			cddb_free_disc_info (gcd->disc_info);
+			gcd->disc_info = NULL;
+		}
+		
+		gtk_widget_set_sensitive (gcd->trackeditor_b, FALSE);
+
+		/* Set the track menu off. */
+		gtk_widget_set_sensitive (gcd->tracks, FALSE);
+
+		/* All buttons can be used when the tray is open,
+		   they should shut the tray. */
+		gtk_widget_set_sensitive (gcd->back_b, TRUE);
+		gtk_widget_set_sensitive (gcd->rewind_b, TRUE);
+		gtk_widget_set_sensitive (gcd->play_b, TRUE);
+		gtk_widget_set_sensitive (gcd->stop_b, TRUE);
+		gtk_widget_set_sensitive (gcd->ffwd_b, TRUE);
+		gtk_widget_set_sensitive (gcd->next_b, TRUE);
+		gtk_widget_set_sensitive (gcd->eject_b, TRUE);
+
 		cd_display_clear (CD_DISPLAY (gcd->display));
-		cd_display_set_line (CD_DISPLAY (gcd->display), CD_DISPLAY_LINE_TIME, "Drive open");
+		cd_display_set_line (CD_DISPLAY (gcd->display), CD_DISPLAY_LINE_TIME, _("Drive open"));
 		gnome_cd_set_window_title (gcd, NULL, NULL);
 		break;
 
 	default:
+		if (gcd->disc_info != NULL) {
+			cddb_free_disc_info (gcd->disc_info);
+			gcd->disc_info = NULL;
+		}
+
+		gtk_widget_set_sensitive (gcd->trackeditor_b, FALSE);
 		cd_display_clear (CD_DISPLAY (gcd->display));
-		cd_display_set_line (CD_DISPLAY (gcd->display), CD_DISPLAY_LINE_TIME, "Drive Error");
+		cd_display_set_line (CD_DISPLAY (gcd->display), CD_DISPLAY_LINE_TIME, _("Drive Error"));
 		gnome_cd_set_window_title (gcd, NULL, NULL);
 		break;
 	}
 
-#ifdef DEBUG
-	g_print ("Status changed\n"
-		 "Device: %d\nAudio %d\n"
-		 "Track: %d (%d:%d:%d)\n"
-		 "--------------------\n",
-		 status->cd, status->audio,
-		 status->track, status->relative.minute,
-		 status->relative.second, status->relative.frame);
-#endif
 	if (gcd->last_status != NULL) {
 		g_free (gcd->last_status);
 	}
@@ -572,3 +770,118 @@ playmode_changed_cb (GtkWidget *display,
 	gcd->cdrom->playmode = mode;
 }
 
+void
+open_preferences (GtkWidget *widget,
+		  GnomeCD *gcd)
+{
+	static GtkWidget *dialog = NULL;
+
+	if (dialog == NULL) {
+		dialog = preferences_dialog_show (gcd, FALSE);
+		g_signal_connect (G_OBJECT (dialog), "destroy",
+				  G_CALLBACK (gtk_widget_destroyed), &dialog);
+	} else {
+		gdk_window_show (dialog->window);
+		gdk_window_raise (dialog->window);
+	}
+}
+
+#define CDDBSLAVE_TRACK_EDITOR_IID "OAFIID:GNOME_Media_CDDBSlave2_TrackEditor"
+
+static void
+restart_track_editor (gpointer data)
+{
+	CORBA_Environment ev;
+	
+	CORBA_exception_init (&ev);
+	track_editor = bonobo_activation_activate_from_id (CDDBSLAVE_TRACK_EDITOR_IID, 0, NULL, &ev);
+	if (BONOBO_EX (&ev)) {
+		g_warning ("Could not reactivate track editor.\n%s", CORBA_exception_id (&ev));
+		track_editor = CORBA_OBJECT_NIL;
+		CORBA_exception_free (&ev);
+
+		return;
+	}
+
+	ORBit_small_listen_for_broken (track_editor, G_CALLBACK (restart_track_editor), &ev);
+	CORBA_exception_free (&ev);
+}
+
+void
+destroy_track_editor (void)
+{
+	if (track_editor != CORBA_OBJECT_NIL) {
+		ORBit_small_unlisten_for_broken (track_editor, G_CALLBACK (restart_track_editor));
+		bonobo_object_release_unref (track_editor, NULL);
+	}
+}
+
+void
+open_track_editor (GtkWidget *widget,
+		   GnomeCD *gcd)
+{
+	GError *error;
+	GnomeCDRomCDDBData *data;
+	CORBA_Environment ev;
+	char *discid;
+	
+	if (gnome_cdrom_get_cddb_data (gcd->cdrom, &data, &error) == FALSE) {
+		gcd_warning ("gnome_cdrom_get_cddb_data returned FALSE: %s", error);
+		g_error_free (error);
+		return;
+	}
+	
+	CORBA_exception_init (&ev);
+	if (track_editor == CORBA_OBJECT_NIL) {
+		track_editor = bonobo_activation_activate_from_id (CDDBSLAVE_TRACK_EDITOR_IID, 0, NULL, &ev);
+		if (BONOBO_EX (&ev)) {
+			/* FIXME: Should be an error dialog */
+			g_warning ("Could not activate track editor.\n%s",
+				   CORBA_exception_id (&ev));
+			CORBA_exception_free (&ev);
+			
+			return;
+		}
+		
+		if (track_editor == CORBA_OBJECT_NIL) {
+			/* FIXME: Should be an error dialog */
+			g_warning ("Could not start track editor.");
+			
+			return;
+		}
+
+		/* Listen for the trackeditor dying on us,
+		   then restart it */
+		ORBit_small_listen_for_broken (track_editor, G_CALLBACK (restart_track_editor), &ev);
+	}
+
+	if (data != NULL) {
+		discid = g_strdup_printf ("%08lx", (gulong) data->discid);
+		GNOME_Media_CDDBTrackEditor_setDiscID (track_editor, discid, &ev);
+		g_free (discid);
+	}
+	
+	GNOME_Media_CDDBTrackEditor_showWindow (track_editor, &ev);
+	if (BONOBO_EX (&ev)) {
+		/* FIXME: Should be an error dialog */
+		CORBA_exception_free (&ev);
+		bonobo_object_release_unref (track_editor, NULL);
+		return;
+	}
+	CORBA_exception_free (&ev);
+}
+
+void
+volume_changed (GtkRange *range,
+		GnomeCD *gcd)
+{
+	double volume;
+	GError *error;
+
+	volume = gtk_range_get_value (range);
+
+	if (gnome_cdrom_set_volume (gcd->cdrom, (int) volume, &error) == FALSE) {
+		gcd_warning ("Error setting volume: %s", error);
+		g_error_free (error);
+	}
+}
